import json
from typing import Literal, Never, overload
from pydantic import BaseModel
from .llm import LLM
from ..types import LlmRequestParams, UserMessage


class OneTurn[Input=str, Output: BaseModel=Never]:
    def __init__(self,
                 llm: LLM,
                 instruction: str,
                 output: Literal["text"] | type[Output] = "text",
                 validate: bool = False,
                 ):
        self._llm = llm
        self._instruction = instruction
        self._output = output
        self._validate = validate
        if validate and output == "text":
            raise ValueError("validate=True requires output to be a pydantic model")

    def _create_request(self, input: Input) -> LlmRequestParams:
        formated_input = self.format_input(input)
        return LlmRequestParams(
            messages=[UserMessage(content=formated_input)],
            instructions=self._instruction,
            output=self._output,
            tool_choice="none",
        )

    def format_input(self, input: Input) -> str:
        return str(input)

    @overload
    async def __call__(self: OneTurn[Input, Never], input: Input) -> str: ...
    @overload
    async def __call__(self: OneTurn[Input, Output], input: Input) -> Output: ...
    async def __call__(self, input: Input) -> str | Output:
        request = self._create_request(input)
        response = await self._llm.generate_text(request)
        if response.content is None:
            raise ValueError("Expected JSON output but got None")

        if self._output == "text":
            return response.content

        if self._validate:
            return self._output.model_validate_json(response.content)
        else:
            data = json.loads(response.content)
            return self._output.model_construct(**data)



import json
import inspect
from functools import singledispatch
from typing import Any, Callable, assert_never, cast
from types import FunctionType, MethodType
from ..types.tool import ToolDef, ToolLike


def _arguments_normalizer(arguments: str | dict) -> dict:
    if isinstance(arguments, str):
        if len(arguments.strip()) == 0:
            return {}
        parsed = json.loads(arguments)
        return cast(dict, parsed)
    elif isinstance(arguments, dict):
        return arguments
    else:
        assert_never(arguments)

def _result_normalizer(result: Any) -> str:
    if isinstance(result, str):
        return result
    return json.dumps(result, ensure_ascii=False)

@singledispatch
async def execute_tool(tool: ToolLike, arguments: str | dict) -> str:
    """
    Raises:
        ValueError: If the tool type is not supported.
        JSONDecodeError: If the arguments is a string but not valid JSON.
    """
    raise ValueError(f"Invalid tool type: {type(tool)}")

@execute_tool.register(FunctionType)
@execute_tool.register(MethodType)
async def _(toolfn: Callable, arguments: str | dict) -> str:
    arguments = _arguments_normalizer(arguments)
    result = (await toolfn(**arguments)
             if inspect.iscoroutinefunction(toolfn)
             else toolfn(**arguments))
    return _result_normalizer(result)

@execute_tool.register(ToolDef)
async def _(tooldef: ToolDef, arguments: str | dict) -> str:
    arguments = _arguments_normalizer(arguments)
    result = (await tooldef.execute(**arguments)
             if inspect.iscoroutinefunction(tooldef.execute)
             else tooldef.execute(**arguments))
    return _result_normalizer(result)



import asyncio
import json
from typing import TYPE_CHECKING, Any, Callable
from .execute import execute_tool
from .utils import get_tool_name
from ..types import LlmToolException, ToolArgumentDecodeError, ToolExecutionError
from ..logger import logger

if TYPE_CHECKING:
    from ..types import ToolLike


type ExceptionHandler[E: LlmToolException] = Callable[[E], str]

class ToolExceptionHandlerManager:
    def __init__(self):
        self._handlers: dict[type[LlmToolException], ExceptionHandler[Any]] = {}

    def register[E: LlmToolException](self, exception_type: type[E]):
        def decorator(handler: ExceptionHandler[E]) -> ExceptionHandler[E]:
            self.set_handler(exception_type, handler)
            return handler
        return decorator

    def set_handler[E: LlmToolException](self, exception_type: type[E], handler: ExceptionHandler[E]):
        self._handlers[exception_type] = handler

    def get_handler[E: LlmToolException](self, exception_type: type[E]) -> ExceptionHandler[E] | None:
        return self._handlers.get(exception_type)

    def handle(self, e: LlmToolException) -> str:
        def find_best_handler[E: LlmToolException](exc_type: type[E]) -> ExceptionHandler[E] | None:
            for cls in exc_type.__mro__:
                if cls in self._handlers:
                    return self._handlers[cls]
            return None

        # Searches the MRO of the exception type to make sure the subclasses of
        # the registered exception type can also be handled.
        handler = find_best_handler(type(e))
        if handler is None:
            logger.warning(f"Unhandled tool exception: {type(e).__name__}", exc_info=e)
            return f"Unhandled tool exception | {type(e).__name__}: {e}"
        return handler(e)

class ToolCallExecutor:
    def __init__(self):
        self._exception_handler = ToolExceptionHandlerManager()

    @property
    def exception_handler(self) -> ToolExceptionHandlerManager:
        return self._exception_handler

    async def execute(self,
                      tool: ToolLike,
                      arguments: str | dict) -> tuple[str | None, str | None]:
        """
        Returns:
            A tuple of (result, error)
        """
        result, error = None, None
        try:
            result = await execute_tool(tool, arguments)
        except json.JSONDecodeError as e:
            assert type(arguments) is str
            _error = ToolArgumentDecodeError(get_tool_name(tool), arguments, e)
            error = self._exception_handler.handle(_error)
        except Exception as e:
            _error = ToolExecutionError(tool, arguments, e)
            error = self._exception_handler.handle(_error)
        return result, error

    def execute_sync(self,
                     tool: ToolLike,
                     arguments: str | dict
                     ) -> tuple[str | None, str | None]:
        """
        Synchronous wrapper of `execute`.
        """
        return asyncio.run(self.execute(tool, arguments))

__all__ = [
    "ToolCallExecutor"
]



import zipfile
import frontmatter
from binaryornot.helpers import is_binary_string
from dataclasses import dataclass, field
from pathlib import PurePosixPath
from typing import Any, Literal, cast
from ..types.exceptions import InvalidSkillArchiveError


ZipPath = PurePosixPath

@dataclass
class SkillMd:
    name: str
    description: str
    content: str

    license: str | None = None
    compatibility: str | None = None
    allowed_tools: str | None = None
    metadata: dict[str, Any] = field(default_factory=dict)

class SkillParser:
    SKILL_MD_NAME = "skill.md"

    @staticmethod
    def find_skill_root(paths: list[ZipPath]) -> ZipPath | None:
        """
        According to the specification, the skill root should be the first-level directory that contains the SKILL.md file.
        If there is no such directory, the root is the archive root.
        """
        for path in paths:
            parts = path.parts
            if len(parts) == 2 and parts[1].lower() == SkillParser.SKILL_MD_NAME:
                return path.parent
            if len(parts) == 1 and path.name.lower() == SkillParser.SKILL_MD_NAME:
                return path.parent
        return None

    @staticmethod
    def find_skill_md(paths: list[ZipPath], root: ZipPath) -> ZipPath | None:
        for path in paths:
            if path.name.lower() == SkillParser.SKILL_MD_NAME and path.parent == root:
                return path
        return None

    @staticmethod
    def parse_skill_md(text: str) -> SkillMd:
        def resolve_optional_str(value: Any) -> str | None:
            if value is None:
                return None
            if isinstance(value, str):
                return value
            return str(value)

        result = frontmatter.loads(text)
        return SkillMd(
            name=str(result["name"]),
            description=str(result["description"]),
            content=result.content,

            license=resolve_optional_str(result.get("license")),
            compatibility=resolve_optional_str(result.get("compatibility")),
            allowed_tools=resolve_optional_str(result.get("allowed-tools")),
            metadata=cast(dict[str, Any], result.get("metadata", {})),
        )

@dataclass
class SkillResource:
    relative: str
    type: Literal["text", "binary"]
    content: str | bytes

    @classmethod
    def from_bytes(cls, relative: ZipPath, content: bytes) -> SkillResource:
        is_binary = is_binary_string(content)
        type = "binary" if is_binary else "text"
        content_resolved = content if is_binary else content.decode("utf-8-sig", errors="replace")
        return cls(
            relative=str(relative),
            type=type,
            content=content_resolved,
        )

@dataclass
class Skill(SkillMd):
    resources: list[SkillResource] = field(default_factory=list)

    @classmethod
    def from_zip(cls, zip_file: zipfile.ZipFile) -> Skill:
        names = zip_file.namelist()
        paths = [ZipPath(name) for name in names]
        
        skill_root = SkillParser.find_skill_root(paths)
        if skill_root is None:
            raise InvalidSkillArchiveError("Skill root not found")

        skill_md = SkillParser.find_skill_md(paths, skill_root)
        if skill_md is None:
            raise InvalidSkillArchiveError("SKILL.md not found")

        skill_md_text = zip_file.read(str(skill_md)).decode("utf-8-sig", errors="replace")
        skill = SkillParser.parse_skill_md(skill_md_text)

        resources: list[SkillResource] = []
        for info in zip_file.infolist():
            if info.is_dir(): continue
            info_path = ZipPath(info.filename)
            if info_path == skill_md: continue # skip SKILL.md

            relative = info_path.relative_to(skill_root)
            content_bytes = zip_file.read(info.filename)
            resources.append(SkillResource.from_bytes(relative, content_bytes))

        return cls(
            **skill.__dict__,
            resources=resources,
        )

__all__ = [
    "SkillResource",
    "Skill",
]



import asyncio
import httpx
import webbrowser
from typing import Any, NamedTuple, override
from mcp import ClientSession
from mcp.client.auth import OAuthClientProvider
from mcp.client.streamable_http import streamable_http_client
from mcp.shared.auth import OAuthClientMetadata
from pydantic import AnyUrl, BaseModel, ConfigDict, PrivateAttr
from .oauth_server import LocalOAuthServer, OAuthCode, TokenStorage, InMemoryTokenStorage
from .base_mcp_client import McpClient, Tool, ToolResult, McpSessionNotEstablishedError
from ..logger import logger

class OAuthParams(BaseModel):
    oauth_scopes: list[str] | None
    oauth_timeout: int = 120
    _oauth_token_storage: TokenStorage = PrivateAttr(default_factory=InMemoryTokenStorage)

class RemoteServerParams(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    url: str
    bearer_token: str | None
    oauth_params: OAuthParams | None
    http_headers: dict[str, str] | None

# --- --- --- --- --- ---

class OAuthContext(NamedTuple):
    client: httpx.AsyncClient
    server: LocalOAuthServer

    async def aclose(self):
        async with asyncio.TaskGroup() as tg:
            tg.create_task(self.client.aclose())
            tg.create_task(self.server.stop())

class RemoteMcpClient(McpClient):
    def __init__(self,
                 name: str,
                 params: RemoteServerParams,
                 storage: TokenStorage | None = None):
        self._name = name
        self._params = params
        self._session: ClientSession | None = None
        self._oauth_context: OAuthContext | None = self._init_oauth()
        if self._params.oauth_params is not None and storage is not None:
            self._params.oauth_params._oauth_token_storage = storage

        self._run_task: asyncio.Task | None = None
        self._connect_error: BaseException | None = None
        self._ready_event = asyncio.Event()
        self._disconnect_event = asyncio.Event()

    @property
    @override
    def name(self) -> str:
        return self._name

    def _init_http_headers(self) -> dict[str, str] | None:
        if self._params.http_headers is None and self._params.bearer_token is None:
            return None
        headers = {}
        if self._params.http_headers is not None:
            headers.update(self._params.http_headers)
        if self._params.bearer_token is not None:
            headers["Authorization"] = f"Bearer {self._params.bearer_token}"
        return headers

    def _init_oauth(self) -> OAuthContext | None:
        if self._params.oauth_params is None:
            return None

        oauth_params = self._params.oauth_params

        server = LocalOAuthServer(timeout=oauth_params.oauth_timeout)
        scopes = None
        if oauth_params.oauth_scopes is not None:
            scopes = " ".join(oauth_params.oauth_scopes)

        client_provider = OAuthClientProvider(
            server_url=self._params.url,
            client_metadata=OAuthClientMetadata(
                client_name=self._name,
                redirect_uris=[AnyUrl(server.callback_url)],
                grant_types=["authorization_code", "refresh_token"],
                response_types=["code"],
                scope=scopes,
                token_endpoint_auth_method="none",
            ),
            storage=oauth_params._oauth_token_storage,
            redirect_handler=self._handle_redirect,
            callback_handler=self._handle_oauth_callback,
        )
        client = httpx.AsyncClient(auth=client_provider,
                                   headers=self._init_http_headers(),
                                   follow_redirects=True)
        return OAuthContext(client, server)

    async def _handle_redirect(self, url: str) -> None:
        logger.info("[OAuth] Authentication required, opening browser...")
        logger.info(f"[OAuth] If browser does not open automatically, copy and open the following link: \n{url}\n")
        try:
            webbrowser.open(url)
        except Exception as e:
            logger.error("[OAuth] Not able to open browser", exc_info=e)

    async def _handle_oauth_callback(self) -> OAuthCode:
        if self._oauth_context is None:
            raise ValueError("OAuth context not initialized")
        return await self._oauth_context.server.wait_for_code()

    async def _run(self):
        custum_http_client: httpx.AsyncClient | None = None
        if self._oauth_context:
            http_client = self._oauth_context.client
            await self._oauth_context.server.start()
        else:
            http_client = httpx.AsyncClient(headers=self._init_http_headers(), follow_redirects=True)
            custum_http_client = http_client

        try:
            async with streamable_http_client(self._params.url, http_client=http_client) as (read_stream, write_stream, _):
                async with ClientSession(read_stream, write_stream) as session:
                    await session.initialize()
                    self._session = session
                    self._ready_event.set()
                    await self._disconnect_event.wait()
        except BaseException as e:
            self._connect_error = e
            self._ready_event.set()
        finally:
            self._session = None
            if custum_http_client:
                await custum_http_client.aclose()

    @override
    async def connect(self):
        self._run_task = asyncio.create_task(self._run())
        await self._ready_event.wait()
        if self._connect_error:
            raise self._connect_error

    @override
    async def list_tools(self) -> list[Tool]:
        if not self._session:
            raise McpSessionNotEstablishedError()

        result = await self._session.list_tools()
        return result.tools

    @override
    async def call_tool(
        self, tool_name: str, arguments: dict[str, Any] | None = None
    ) -> ToolResult:
        if not self._session:
            raise McpSessionNotEstablishedError()

        response = await self._session.call_tool(tool_name, arguments)
        return ToolResult(response.isError, response.content)

    @override
    async def disconnect(self):
        if self._disconnect_event:
            self._disconnect_event.set()

        if self._run_task and not self._run_task.done():
            try:
                await self._run_task
            except Exception:
                pass
        self._run_task = None

        if self._oauth_context:
            try:
                await self._oauth_context.aclose()
            except* Exception: pass



import json
from typing import Literal, cast, override
from anthropic import AsyncAnthropic, ParsedMessageStreamEvent
from anthropic.types import ImageBlockParam, Message, MessageParam, TextBlockParam, ToolChoiceAnyParam, ToolChoiceAutoParam, ToolChoiceNoneParam, ToolParam, ToolResultBlockParam, ToolUseBlock, ToolUseBlockParam
from anthropic.types.message_create_params import MessageCreateParamsBase, MessageCreateParamsNonStreaming
from anthropic.types.text_block_param import TextBlockParam
from anthropic.types.tool_param import InputSchema
from pydantic import BaseModel
from .base_provider import BaseMessageParser, BaseParamParser, BaseProvider
from .utils import StreamMessageCollector, StrictInlineJsonSchema
from ..tool.prepare import prepare_tools
from ..types import (
    LlmRequestParams,
    Attachment, ImageAttachment, AttachmentTypeNotSupportedError,
    BaseMessage, SystemMessage, UserMessage, ToolMessage, AssistantMessage,
    TextChunkEvent, ToolCallChunkEvent, UsageChunkEvent, AssistantMessageEvent,
)


class AnthropicProviderMessageParser(BaseMessageParser[
    ParsedMessageStreamEvent,
    Message,
    MessageParam,
]):
    @staticmethod
    def _attachment_to_content_part(attachment: Attachment) -> ImageBlockParam:
        match attachment:
            case ImageAttachment() if attachment.source.type == "url":
                # Anthropic supports URL sources for images
                return ImageBlockParam(
                    type="image",
                    source={"type": "url", "url": attachment.source.url},
                )
            case ImageAttachment() if attachment.source.type == "base64":
                if attachment.source.mime_type not in ["image/jpeg", "image/png", "image/gif", "image/webp"]:
                    raise AttachmentTypeNotSupportedError(attachment.source.mime_type)
                media_type = cast(Literal["image/jpeg", "image/png", "image/gif", "image/webp"], attachment.source.mime_type)
                return ImageBlockParam(
                    type="image",
                    source={
                        "type": "base64",
                        "media_type": media_type,
                        "data": attachment.source.data
                    },
                )
            case _:
                # Anthropic does not support audio input via the messages API
                raise AttachmentTypeNotSupportedError(attachment.type)

    @staticmethod
    def normalize_chunk(chunk: ParsedMessageStreamEvent) -> list[TextChunkEvent | ToolCallChunkEvent | UsageChunkEvent]:
        result: list[TextChunkEvent | ToolCallChunkEvent | UsageChunkEvent] = []
        match chunk.type:
            case "text":
                result.append(TextChunkEvent(chunk.text))
            case "content_block_stop" if chunk.content_block.type == "tool_use":
                result.append(ToolCallChunkEvent(
                    id=chunk.content_block.id,
                    name=chunk.content_block.name,
                    arguments=json.dumps(chunk.content_block.input),
                    index=chunk.index))
        return result

    @override
    @staticmethod
    def to_message(response: Message) -> AssistantMessage:
        content_text: str | None = None
        reasoning_content: str | None = None
        tool_calls: list[AssistantMessage.ToolCall] | None = None
 
        for block in response.content:
            if block.type == "text":
                content_text = block.text
            elif block.type == "thinking":
                reasoning_content = block.thinking
            elif block.type == "tool_use":
                if tool_calls is None:
                    tool_calls = []
                tool_calls.append(AssistantMessage.ToolCall(
                    id=block.id,
                    name=block.name,
                    arguments=block.input,
                ))
 
        usage = response.usage
        return AssistantMessage(
            content=content_text,
            reasoning_content=reasoning_content,
            tool_calls=tool_calls,
            usage=AssistantMessage.Usage(
                input_tokens=usage.input_tokens,
                output_tokens=usage.output_tokens,
                total_tokens=usage.input_tokens + usage.output_tokens,
            ) if usage else None,
        )

    @override
    @staticmethod
    def from_message(message: BaseMessage) -> MessageParam:
        match message:
            case UserMessage() if message.attachments is None:
                return MessageParam(role="user", content=message.content)

            case UserMessage() if message.attachments is not None:
                parts: list[TextBlockParam | ImageBlockParam] = [
                    TextBlockParam(type="text", text=message.content)
                ]
                for attachment in message.attachments:
                    parts.append(
                        AnthropicProviderMessageParser._attachment_to_content_part(attachment)
                    )
                return MessageParam(role="user", content=parts)
 
            case AssistantMessage():
                content_blocks = []
                if message.content is not None:
                    content_blocks.append(
                        TextBlockParam(type="text", text=message.content)
                    )
                if message.tool_calls is not None:
                    for tc in message.tool_calls:
                        content_blocks.append(ToolUseBlockParam(
                            type="tool_use",
                            id=tc.id,
                            name=tc.name,
                            input=tc.arguments,
                        ))
 
                # Anthropic rejects an assistant message with an empty content array,
                # fall back to an empty string if there is nothing.
                if len(content_blocks) == 0:
                    content_blocks.append(TextBlockParam(type="text", text=""))

                return MessageParam(role="assistant", content=content_blocks)
 
            case ToolMessage() as message:
                # Anthropic tool results travel as a *user* message containing a tool_result block.
                return MessageParam(
                    role="user",
                    content=[ToolResultBlockParam(
                        type="tool_result",
                        tool_use_id=message.call_id,
                        content=message.content,
                    )],
                )
 
            case SystemMessage():
                raise ValueError(
                    "SystemMessage must be passed as the `system` parameter, "
                    "not in the messages array."
                )
            case _:
                raise NotImplementedError(
                    f"Unsupported message type: {type(message)}"
                )

class AnthropicProviderParamParser(BaseParamParser[
    MessageCreateParamsNonStreaming,
    MessageCreateParamsBase,
]):
    def _preparse_tools(self, params: LlmRequestParams) -> list[ToolParam] | None:
        extracted_tool_likes = params.extract_tools()
        if extracted_tool_likes is None: return None

        tool_schemas = prepare_tools(extracted_tool_likes)
        return [ToolParam(
            name=tool_schema["name"],
            description=tool_schema["description"],
            input_schema=cast(InputSchema, tool_schema["parameters"]),
        ) for tool_schema in tool_schemas]

    def _preparse_messages(self, params: LlmRequestParams) -> list[MessageParam]:
        transformed_messages: list[MessageParam] = []
        for message in params.messages:
            if (type(message) is ToolMessage and not message.is_complete):
                continue
            parsed_message = self._message_parser.from_message(message)
            transformed_messages.append(parsed_message)
        return transformed_messages

    @override
    def parse_nonstream(self, params: LlmRequestParams) -> MessageCreateParamsNonStreaming:
        base_params = cast(MessageCreateParamsNonStreaming, self.parse_stream(params))
        base_params["stream"] = False
        return base_params

    @override
    def parse_stream(self, params: LlmRequestParams) -> MessageCreateParamsBase:
        assert params.model is not None
        result  = MessageCreateParamsBase(
            model=params.model,
            max_tokens=32767,
            messages=self._preparse_messages(params),
            **(params.extra_args or {})
        )
        if params.instructions is not None:
            result["system"] = params.instructions
        if params.temperature is not None:
            result["temperature"] = params.temperature
        if params.max_tokens is not None:
            result["max_tokens"] = params.max_tokens
        if (tools := self._preparse_tools(params)) is not None:
            result["tools"] = tools
        if params.tool_choice is not None:
            if "tools" not in result:
                # Anthropic SDK requires tools to be set if tool_choice is not "none"
                params.tool_choice = "none"
            match params.tool_choice:
                case "auto":     result["tool_choice"] = ToolChoiceAutoParam(type="auto")
                case "required": result["tool_choice"] = ToolChoiceAnyParam(type="any")
                case "none":     result["tool_choice"] = ToolChoiceNoneParam(type="none")
        if params.output is not None and params.output != "text":
            if params.output == "json":
                result["output_config"] = {"format": {
                    "type": "json_schema",
                    "schema": {
                        "type": "object",
                        "properties": {},
                        "additionalProperties": True
                    }
                }}
            if isinstance(params.output, BaseModel):
                result["output_config"] = {"format": {
                    "type": "json_schema",
                    "schema": params.output.model_json_schema(schema_generator=StrictInlineJsonSchema),
                }}
        return result

class AnthropicProvider(BaseProvider):
    def __init__(self, base_url: str, api_key: str):
        self._client = AsyncAnthropic(
            base_url=base_url,
            api_key=api_key,
        )
        self._message_parser = AnthropicProviderMessageParser()
        self._param_parser = AnthropicProviderParamParser(self._message_parser)

    @override
    async def list_models(self) -> list[str]:
        models = await self._client.models.list()
        return [model.id for model in models.data]

    @override
    async def request_nonstream(self, params: LlmRequestParams):
        parsed = self._param_parser.parse_nonstream(params)
        response = await self._client.messages.create(
            **parsed,
            timeout=params.timeout_sec,
            extra_headers=params.headers,
        )
        return self._message_parser.to_message(response)

    @override
    async def request_stream(self, params: LlmRequestParams):
        parsed = self._param_parser.parse_stream(params)
        message_collector = StreamMessageCollector()

        async with self._client.messages.stream(
            **parsed,
            timeout=params.timeout_sec,
            extra_headers=params.headers,
        ) as stream:
            async for event in stream:
                normalized_chunks = self._message_parser.normalize_chunk(event)
                for chunk in normalized_chunks:
                    yield chunk
                    message_collector.collect(chunk)

            message = await stream.get_final_message()
            yield UsageChunkEvent(
                input_tokens=message.usage.input_tokens,
                output_tokens=message.usage.output_tokens,
                total_tokens=message.usage.input_tokens + message.usage.output_tokens,
            )

            full_message = message_collector.get_message()
            full_message.usage = AssistantMessage.Usage(
                input_tokens=message.usage.input_tokens,
                output_tokens=message.usage.output_tokens,
                total_tokens=message.usage.input_tokens + message.usage.output_tokens,
            )
            yield AssistantMessageEvent(message=full_message)



import asyncio
from typing import Any, override
from mcp import ClientSession, StdioServerParameters as StdioServerParams
from mcp.client.stdio import stdio_client
from .base_mcp_client import McpClient, Tool, ToolResult, McpSessionNotEstablishedError

class LocalServerParams(StdioServerParams): ...

class LocalMcpClient(McpClient):
    def __init__(self, name: str, params: LocalServerParams):
        self._name: str = name
        self._params: LocalServerParams = params
        self._session: ClientSession | None = None
        self._run_task: asyncio.Task | None = None

        self._connect_error: BaseException | None = None
        self._ready_event = asyncio.Event()
        self._disconnect_event = asyncio.Event()

    async def _run(self):
        try:
            async with stdio_client(self._params) as (read_stream, write_stream):
                async with ClientSession(read_stream, write_stream) as session:
                    await session.initialize()
                    self._session = session
                    self._ready_event.set()
                    await self._disconnect_event.wait()
        except BaseException as e:
            self._connect_error = e
            self._ready_event.set()
        finally:
            self._session = None

    @property
    @override
    def name(self) -> str:
        return self._name

    @override
    async def connect(self):
        self._run_task = asyncio.create_task(self._run())
        await self._ready_event.wait()
        if self._connect_error:
            raise self._connect_error

    @override
    async def list_tools(self) -> list[Tool]:
        if not self._session:
            raise McpSessionNotEstablishedError()

        result = await self._session.list_tools()
        return result.tools

    @override
    async def call_tool(
        self, tool_name: str, arguments: dict[str, Any] | None = None
    ) -> ToolResult:
        if not self._session:
            raise McpSessionNotEstablishedError()

        response = await self._session.call_tool(tool_name, arguments)
        return ToolResult(response.isError, response.content)

    @override
    async def disconnect(self) -> None:
        if self._disconnect_event:
            self._disconnect_event.set()

        if self._run_task and not self._run_task.done():
            try:
                await self._run_task
            except Exception:
                pass
        self._run_task = None



import asyncio
from typing import TYPE_CHECKING
from collections.abc import Generator
from ..providers import LlmProviders

if TYPE_CHECKING:
    from ..providers import BaseProvider
    from ..types import (
        LlmRequestParams, StreamMessageGenerator,
        StreamMessageEvent, AssistantMessage,
    )

class LLM:
    def __init__(self, name: str, provider: BaseProvider):
        self._name = name
        self._provider = provider

    @staticmethod
    def create_provider(provider_type: LlmProviders, base_url: str, api_key: str) -> BaseProvider:
        match provider_type:
            case LlmProviders.OPENAI:
                from ..providers.openai import OpenAIProvider
                return OpenAIProvider(base_url, api_key)
            case LlmProviders.ANTHROPIC:
                from ..providers.anthropic import AnthropicProvider
                return AnthropicProvider(base_url, api_key)
            case _:
                raise ValueError(f"Unsupported provider type: {provider_type}")

    async def generate_text(self, params: LlmRequestParams) -> AssistantMessage:
        params.model = params.model or self._name
        return await self._provider.request_nonstream(params)

    def generate_text_sync(self, params: LlmRequestParams) -> AssistantMessage:
        return asyncio.run(self.generate_text(params))

    async def stream_text(self, params: LlmRequestParams) -> StreamMessageGenerator:
        params.model = params.model or self._name
        async for chunk in self._provider.request_stream(params):
            yield chunk

    def stream_text_sync(self, params: LlmRequestParams) -> Generator[StreamMessageEvent, None, None]:
        with asyncio.Runner() as runner:
            gen = self.stream_text(params)
            while True:
                try:
                    chunk = runner.run(gen.__anext__())
                    yield chunk
                except StopAsyncIteration:
                    break
